from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, Sequence, TYPE_CHECKING

import torch
from aislib.pytorch_modules import Swish
from torch import nn

from eir.models.fusion import default_fuse_features, al_features
from eir.models.layers import MLPResidualBlock
from eir.models.models_base import (
    construct_multi_branches,
    initialize_modules_from_spec,
    create_multi_task_blocks_with_first_adaptor_block,
    construct_blocks,
    get_final_layer,
    merge_module_dicts,
    calculate_module_dict_outputs,
)

if TYPE_CHECKING:
    from eir.train import al_num_outputs_per_target


@dataclass
class MGMoEModelConfig:
    layers: Sequence[int]
    fc_task_dim: int

    split_mlp_num_splits: int
    mg_num_experts: int

    rb_do: float
    fc_do: float


class MGMoEModel(nn.Module):
    def __init__(
        self,
        model_config: MGMoEModelConfig,
        num_outputs_per_target: "al_num_outputs_per_target",
        modules_to_fuse: nn.ModuleDict,
        fusion_callable: al_features = default_fuse_features,
    ):
        super().__init__()

        self.model_config = model_config
        self.num_outputs_per_target = num_outputs_per_target
        self.modules_to_fuse = modules_to_fuse
        self.fusion_callable = fusion_callable

        self.num_chunks = self.model_config.split_mlp_num_splits
        self.num_experts = self.model_config.mg_num_experts

        cur_dim = sum(i.num_out_features for i in self.modules_to_fuse.values())

        self.task_names = sorted(tuple(self.num_outputs_per_target.keys()))
        gate_spec = self.get_gate_spec(
            in_features=cur_dim, out_features=self.num_experts
        )

        self.gates = construct_multi_branches(
            branch_names=self.task_names,
            branch_factory=initialize_modules_from_spec,
            branch_factory_kwargs={"spec": gate_spec},
        )

        expert_names = tuple(f"expert_{i}" for i in range(self.num_experts))
        layer_kwargs = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "full_preactivation": False,
        }
        self.expert_branches = create_multi_task_blocks_with_first_adaptor_block(
            num_blocks=self.model_config.layers[0],
            branch_names=expert_names,
            block_constructor=MLPResidualBlock,
            block_constructor_kwargs=layer_kwargs,
            first_layer_kwargs_overload={
                "full_preactivation": True,
                "in_features": cur_dim,
            },
        )

        task_resblocks_kwargs = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "full_preactivation": False,
        }
        multi_task_branches = construct_multi_branches(
            branch_names=self.task_names,
            branch_factory=construct_blocks,
            branch_factory_kwargs={
                "num_blocks": self.model_config.layers[1],
                "block_constructor": MLPResidualBlock,
                "block_kwargs": task_resblocks_kwargs,
            },
        )

        final_act_spec = self.get_final_act_spec(
            in_features=self.model_config.fc_task_dim, dropout_p=self.model_config.fc_do
        )
        final_act = construct_multi_branches(
            branch_names=self.task_names,
            branch_factory=initialize_modules_from_spec,
            branch_factory_kwargs={"spec": final_act_spec},
        )

        final_layer = get_final_layer(
            in_features=self.model_config.fc_task_dim,
            num_outputs_per_target=self.num_outputs_per_target,
        )

        self.multi_task_branches = merge_module_dicts(
            (multi_task_branches, final_act, final_layer)
        )

        self._init_weights()

    @staticmethod
    def get_gate_spec(in_features: int, out_features: int):

        spec = OrderedDict(
            {
                "gate_fc": (
                    nn.Linear,
                    {
                        "in_features": in_features,
                        "out_features": out_features,
                        "bias": True,
                    },
                ),
                "gate_attention": (nn.Softmax, {"dim": 1}),
            }
        )

        return spec

    @staticmethod
    def get_final_act_spec(in_features: int, dropout_p: float):
        # TODO: Refactor, duplicated from fusion.py

        spec = OrderedDict(
            {
                "bn_final": (nn.BatchNorm1d, {"num_features": in_features}),
                "act_final": (Swish, {}),
                "do_final": (nn.Dropout, {"p": dropout_p}),
            }
        )

        return spec

    def _init_weights(self):
        pass

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        # TODO: Refactor into class for fusion models.
        out = []
        for module in self.modules_to_fuse.values():
            if hasattr(module, "l1_penalized_weights"):
                weight_flat = torch.flatten(module.l1_penalized_weights)
                out.append(weight_flat)

        return torch.cat(out)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:

        out = {}
        for module_name, module in self.modules_to_fuse.items():
            module_input = inputs[module_name]
            out[module] = module(module_input)

        fused_features = default_fuse_features(tuple(out.values()))

        gate_attentions = calculate_module_dict_outputs(
            input_=fused_features, module_dict=self.gates
        )

        expert_outputs = calculate_module_dict_outputs(
            input_=fused_features, module_dict=self.expert_branches
        )

        final_out = {}
        stacked_expert_outputs = torch.stack(list(expert_outputs.values()), dim=2)
        for task_name, task_attention in gate_attentions.items():
            weighted_expert_outputs = (
                task_attention.unsqueeze(1) * stacked_expert_outputs
            )
            weighted_expert_sum = weighted_expert_outputs.sum(dim=2)

            cur_task_branch = self.multi_task_branches[task_name]
            final_out[task_name] = cur_task_branch(weighted_expert_sum)

        return final_out



from dataclasses import dataclass
from typing import Tuple, Dict, List

import torch

from eir.data_load.label_setup import al_target_columns


def get_target_columns_generator(target_columns: al_target_columns) -> Tuple[str, str]:
    for column_type, list_of_cols_of_this_type in target_columns.items():
        for cur_column in list_of_cols_of_this_type:
            yield column_type, cur_column


@dataclass(frozen=True)
class Batch:
    inputs: Dict[str, torch.Tensor]
    target_labels: Dict[str, torch.Tensor]
    ids: List[str]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['eir',
 'eir.data_load',
 'eir.interpretation',
 'eir.models',
 'eir.models.omics',
 'eir.models.tabular',
 'eir.train_utils',
 'eir.visualization']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0',
 'adabelief-pytorch>=0.2.0,<0.3.0',
 'aislib==0.1.5a0',
 'colorama>=0.4.4,<0.5.0',
 'dill>=0.3.3,<0.4.0',
 'hypothesis>=6.14.0,<7.0.0',
 'joblib>=0.17.0,<0.18.0',
 'matplotlib>=3.3.2,<4.0.0',
 'numpy>=1.19.2,<2.0.0',
 'pandas>=1.2.0,<2.0.0',
 'plotly>=4.11.0,<5.0.0',
 'py>=1.9.0,<2.0.0',
 'pytorch-ignite>=0.4.2,<0.5.0',
 'scikit-learn>=0.24.0,<0.25.0',
 'seaborn>=0.11.1,<0.12.0',
 'shap>=0.39.0,<0.40.0',
 'sympy>=1.6.2,<2.0.0',
 'tensorboard>=2.3.0,<3.0.0',
 'torch-optimizer>=0.1.0,<0.2.0',
 'torch==1.9.0',
 'torchvision==0.10.0',
 'tqdm>=4.55.0,<5.0.0']

entry_points = \
{'console_scripts': ['eirpredict = eir.predict:main',
                     'eirtrain = eir.train:main']}

setup_kwargs = {
    'name': 'eir-dl',
    'version': '0.1.6a0',
    'description': '',
    'long_description': None,
    'author': 'Arnor Sigurdsson',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Callable, Dict, TYPE_CHECKING, Sequence

import torch
from aislib.misc_utils import get_logger
from torch import nn

if TYPE_CHECKING:
    from eir.train import DataDimensions

logger = get_logger(__name__)


@dataclass
class LinearModelConfig:
    input_name: str
    data_dimensions: "DataDimensions"
    target_cat_columns: Sequence[str]
    target_con_columns: Sequence[str]


class LinearModel(nn.Module):
    def __init__(self, model_config: LinearModelConfig):
        super().__init__()

        self.model_config = model_config

        self.fc_1 = nn.Linear(self.fc_1_in_features, 1)
        self.act = self._get_act()

        self.output_parser = self._get_output_parser()

    @property
    def fc_1_in_features(self) -> int:
        return self.model_config.data_dimensions.num_elements()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.fc_1.weight

    @property
    def num_out_features(self) -> int:
        if self.model_config.target_cat_columns:
            return 2
        return 1

    def _get_act(self) -> Callable[[torch.Tensor], torch.Tensor]:
        if self.model_config.target_cat_columns:
            logger.info(
                "Using logistic regression model on categorical column: %s.",
                self.model_config.target_cat_columns,
            )
            return nn.Sigmoid()

        # no activation for linear regression
        elif self.model_config.target_con_columns:
            logger.info(
                "Using linear regression model on continuous column: %s.",
                self.model_config.target_con_columns,
            )
            return lambda x: x

        raise ValueError()

    def _get_output_parser(self) -> Callable[[torch.Tensor], Dict[str, torch.Tensor]]:
        def _parse_categorical(out: torch.Tensor) -> Dict[str, torch.Tensor]:
            # we create a 2D output from 1D for compatibility with visualization funcs
            out = torch.cat(((1 - out[:, 0]).unsqueeze(1), out), 1)
            out = {self.model_config.target_cat_columns[0]: out}
            return out

        def _parse_continuous(out: torch.Tensor) -> Dict[str, torch.Tensor]:
            out = {self.model_config.target_con_columns[0]: out}
            return out

        if self.model_config.target_cat_columns:
            return _parse_categorical
        return _parse_continuous

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        genotype = inputs[self.model_config.input_name]
        out = genotype.view(genotype.shape[0], -1)

        tabular = inputs.get("tabular_cl_args", None)
        if tabular is not None:
            out = torch.cat((out, tabular), dim=1)

        out = self.fc_1(out)
        out = self.act(out)

        out = self.output_parser(out)

        return out



from argparse import Namespace
from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import (
    List,
    Dict,
    Union,
    Tuple,
    Callable,
    Iterable,
    Sequence,
    Generator,
    DefaultDict,
    Any,
    TYPE_CHECKING,
)

import numpy as np
import torch
from aislib.misc_utils import get_logger
from torch.utils.data import Dataset
from tqdm import tqdm

from eir.data_load.data_augmentation import make_random_omics_columns_missing
from eir.data_load.label_setup import (
    al_label_dict,
    al_target_columns,
    get_array_path_iterator,
    Labels,
)
from eir.data_load.label_setup import merge_target_columns

if TYPE_CHECKING:
    from eir.train import DataDimensions

logger = get_logger(name=__name__, tqdm_compatible=True)

# Type Aliases
al_datasets = Union["MemoryDataset", "DiskDataset"]
# embeddings --> remain str, cat targets --> int, con extra/target --> float
al_sample_labels_transformed_all = Dict[str, Union[int, str, float]]
al_sample_label_dict_target = Dict[str, Union[int, float]]
al_sample_label_dict_extra = Dict[str, Union[str, float]]
al_all_labels = Dict[
    str, Union[al_sample_label_dict_target, al_sample_label_dict_extra]
]
al_inputs = Union[Dict[str, torch.Tensor], Dict[str, Any]]
al_getitem_return = Tuple[Dict[str, torch.Tensor], al_sample_label_dict_target, str]


def set_up_datasets_from_cl_args(
    cl_args: Namespace,
    target_labels: Labels,
    data_dimensions: Dict[str, "DataDimensions"],
    tabular_inputs_labels: Union[Labels, None] = None,
) -> Tuple[al_datasets, al_datasets]:

    dataset_class = MemoryDataset if cl_args.memory_dataset else DiskDataset

    train_kwargs = construct_default_dataset_kwargs_from_cl_args(
        cl_args=cl_args,
        target_labels_dict=target_labels.train_labels,
        data_dimensions=data_dimensions,
        tabular_labels_dict=tabular_inputs_labels.train_labels,
        na_augment=True,
    )

    valid_kwargs = construct_default_dataset_kwargs_from_cl_args(
        cl_args=cl_args,
        target_labels_dict=target_labels.valid_labels,
        data_dimensions=data_dimensions,
        tabular_labels_dict=tabular_inputs_labels.valid_labels,
        na_augment=False,
    )

    train_dataset = dataset_class(**train_kwargs)
    valid_dataset = dataset_class(**valid_kwargs)

    _check_valid_and_train_datasets(
        train_dataset=train_dataset, valid_dataset=valid_dataset
    )

    return train_dataset, valid_dataset


def construct_default_dataset_kwargs_from_cl_args(
    cl_args: Namespace,
    target_labels_dict: Union[None, al_label_dict],
    data_dimensions: Dict[str, "DataDimensions"],
    tabular_labels_dict: Union[None, al_label_dict],
    na_augment: bool,
) -> Dict[str, Any]:

    target_columns = merge_target_columns(
        target_con_columns=cl_args.target_con_columns,
        target_cat_columns=cl_args.target_cat_columns,
    )

    data_sources = gather_all_data_sources(
        cl_args=cl_args, tabular_labels_dict=tabular_labels_dict
    )

    dataset_kwargs = {
        "target_columns": target_columns,
        "data_sources": data_sources,
        "data_dimensions": data_dimensions,
        "target_labels_dict": target_labels_dict,
    }

    if na_augment:
        dataset_kwargs["na_augment"] = (
            cl_args.na_augment_perc,
            cl_args.na_augment_prob,
        )

    return dataset_kwargs


def gather_all_data_sources(
    cl_args: Namespace, tabular_labels_dict: Union[Labels, None]
) -> Dict[str, str]:
    all_sources = {}

    if cl_args.omics_sources and cl_args.omics_names:
        for source_on_disk, name in zip(cl_args.omics_sources, cl_args.omics_names):
            all_sources[name] = source_on_disk

    if tabular_labels_dict:
        all_sources["tabular_cl_args"] = tabular_labels_dict

    return all_sources


def _check_valid_and_train_datasets(
    train_dataset: al_datasets, valid_dataset: al_datasets
) -> None:
    assert len(train_dataset) > len(valid_dataset)
    assert set(valid_dataset.target_labels_dict.keys()).isdisjoint(
        train_dataset.target_labels_dict.keys()
    )


@dataclass
class Sample:
    """
    array: can be path to array or the loaded array itself
    """

    sample_id: str
    inputs: Dict[str, Any]
    target_labels: al_sample_label_dict_target


class DatasetBase(Dataset):
    def __init__(
        self,
        data_sources: Dict[str, Any],
        data_dimensions: Dict,
        target_columns: al_target_columns,
        target_labels_dict: al_label_dict = None,
        na_augment: Tuple[float] = (0.0, 0.0),
    ):
        super().__init__()

        self.data_sources = data_sources
        self.data_dimensions = data_dimensions

        self.samples: Union[List[Sample], None] = None

        self.target_columns = target_columns
        self.target_labels_dict = target_labels_dict if target_labels_dict else {}

        self.na_augment = na_augment

    def init_label_attributes(self):
        if not self.target_columns:
            raise ValueError("Please specify label column name.")

    def set_up_samples(self, omics_hook: Callable = lambda x: x) -> List[Sample]:
        """
        We do an extra filtering step at the end to account for the situation where
        we have a target label file with more samples than there are any inputs
        available for. This is quite likely if we have e.g. pre-split data into
        train/val and test folders.
        """

        def _default_factory() -> Sample:
            return Sample(sample_id="", inputs={}, target_labels={})

        samples = defaultdict(_default_factory)

        ids_to_keep = None
        if self.target_labels_dict:
            samples = _add_target_labels_to_samples(
                target_labels_dict=self.target_labels_dict, samples=samples
            )
            ids_to_keep = self.target_labels_dict.keys()

        for source_name, source_data in self.data_sources.items():

            if source_name.startswith("omics_"):

                samples = _add_file_data_to_samples(
                    source_data=source_data,
                    samples=samples,
                    ids_to_keep=ids_to_keep,
                    file_loading_hook=omics_hook,
                    source_name=source_name,
                )

            elif source_name.startswith("tabular_"):
                samples = _add_tabular_data_to_samples(
                    tabular_dict=source_data,
                    samples=samples,
                    ids_to_keep=ids_to_keep,
                    source_name=source_name,
                )

        if self.target_labels_dict:
            num_samples_raw = len(samples)
            samples = list(i for i in samples.values() if i.inputs and i.target_labels)
            num_missing = num_samples_raw - len(samples)
            logger.debug(
                "Filtered out %d samples that had no inputs or no target labels.",
                num_missing,
            )

        return samples

    def __getitem__(self, index: int):
        raise NotImplementedError

    def check_samples(self):

        no_ids, no_inputs, no_target_labels = [], [], []

        for s in self.samples:
            if not s.sample_id:
                no_ids.append(s)

            if not s.inputs:
                no_inputs.append(s)

            if self.target_labels_dict:
                if not s.target_labels:
                    no_target_labels.append(s)

        if no_ids:
            raise ValueError(
                f"Expected all observations to have a sample ID associated "
                f"with them, but got {no_ids}."
            )

        if no_inputs:
            raise ValueError(
                f"Expected all observations to have an input associated "
                f"with them, but got {no_inputs}."
            )

        if self.target_labels_dict:
            if no_target_labels:
                raise ValueError(
                    f"Expected all observations to have a label associated "
                    f"with them, but got {no_target_labels}."
                )


def _add_target_labels_to_samples(
    target_labels_dict: al_label_dict, samples: DefaultDict[str, Sample]
) -> DefaultDict[str, Sample]:
    target_label_iterator = tqdm(target_labels_dict.items(), desc="Target Labels")

    for sample_id, sample_target_labels in target_label_iterator:
        _add_id_to_samples(samples=samples, sample_id=sample_id)
        samples[sample_id].target_labels = sample_target_labels

    return samples


def _add_file_data_to_samples(
    source_data: str,
    samples: DefaultDict[str, Sample],
    ids_to_keep: Union[None, Sequence[str]],
    file_loading_hook: Callable,
    source_name: str = "File Data",
) -> DefaultDict[str, Sample]:

    file_data_iterator = get_file_sample_id_iterator_basic(
        data_source=source_data, ids_to_keep=ids_to_keep
    )
    omics_file_iterator_tqdm = tqdm(file_data_iterator, desc=source_name)

    for sample_id, file in omics_file_iterator_tqdm:

        sample_data = file_loading_hook(file)

        samples = _add_id_to_samples(samples=samples, sample_id=sample_id)

        samples[sample_id].inputs[source_name] = sample_data

    return samples


def _add_tabular_data_to_samples(
    tabular_dict: al_label_dict,
    samples: DefaultDict[str, Sample],
    ids_to_keep: Union[None, Sequence[str]],
    source_name: str = "Tabular Data",
) -> DefaultDict[str, Sample]:

    tabular_iterator = tqdm(tabular_dict.items(), desc=source_name)

    for sample_id, tabular_inputs in tabular_iterator:

        if sample_id not in ids_to_keep:
            continue

        samples = _add_id_to_samples(samples=samples, sample_id=sample_id)

        samples[sample_id].inputs[source_name] = tabular_inputs

    return samples


def _add_id_to_samples(
    samples: DefaultDict[str, Sample], sample_id: str
) -> DefaultDict[str, Sample]:
    """
    This kind of weird function is used because in some cases, we cannot expect the
    target labels to have added samples, because we could be predicting on completely
    unknown samples without any target label data.

    Hence, we might have sparse modular data available for the samples, e.g. only omics
    for some samples, but only tabular data for others. So we want to ensure that the
    data is filled in.
    """
    if not samples[sample_id].sample_id:
        samples[sample_id].sample_id = sample_id
    else:
        assert samples[sample_id].sample_id == sample_id

    return samples


def get_file_sample_id_iterator_basic(
    data_source: str,
    ids_to_keep: Union[None, Sequence[str]],
) -> Generator[Tuple[Any, str], None, None]:

    base_file_iterator = get_array_path_iterator(
        data_source=Path(data_source), validate=False
    )

    for file in base_file_iterator:
        sample_id = file.stem

        if ids_to_keep:
            if sample_id in ids_to_keep:
                yield sample_id, file
        else:
            yield sample_id, file


def get_file_sample_id_iterator(
    data_source: str, ids_to_keep: Union[None, Sequence[str]]
) -> Generator[Tuple[Any, str], None, None]:
    def _id_from_filename(file: Path) -> str:
        return file.stem

    def _filter_ids_callable(item, sample_id):
        if sample_id in ids_to_keep:
            return True
        return False

    base_file_iterator = get_array_path_iterator(
        data_source=Path(data_source), validate=False
    )

    sample_id_and_file_iterator = _get_sample_id_data_iterator(
        base_iterator=base_file_iterator, id_callable=_id_from_filename
    )

    if ids_to_keep:
        final_iterator = _get_filter_iterator(
            base_iterator=sample_id_and_file_iterator,
            filter_callable=_filter_ids_callable,
        )
    else:
        final_iterator = sample_id_and_file_iterator

    yield from final_iterator


def _get_sample_id_data_iterator(
    base_iterator, id_callable: Callable
) -> Generator[Tuple[Any, str], None, None]:
    for item in base_iterator:
        sample_id = id_callable(item)
        yield item, sample_id


def _get_filter_iterator(base_iterator, filter_callable) -> Generator[Any, None, None]:
    for item in base_iterator:
        if filter_callable(*item):
            yield item


class MemoryDataset(DatasetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.target_labels_dict:
            self.init_label_attributes()

        self.samples = self.set_up_samples(omics_hook=self._mem_sample_loader)
        self.check_samples()

    @staticmethod
    def _mem_sample_loader(sample_fpath: Union[str, Path]) -> torch.ByteTensor:
        """
        A small hook to actually load the arrays into `self.samples` instead of just
        pointing to filenames.
        """
        array = np.load(sample_fpath)

        array = array.astype(np.uint8)
        tensor = torch.from_numpy(array).unsqueeze(0)
        return tensor

    def __getitem__(self, index: int) -> al_getitem_return:
        sample = self.samples[index]

        inputs_prepared = copy(sample.inputs)
        inputs_prepared = prepare_inputs_memory(
            inputs=inputs_prepared,
            na_augment_perc=self.na_augment[0],
            na_augment_prob=self.na_augment[1],
        )

        inputs_final = impute_missing_modalities_wrapper(
            inputs=inputs_prepared, data_dimensions=self.data_dimensions
        )

        target_labels = sample.target_labels
        sample_id = sample.sample_id

        return inputs_final, target_labels, sample_id

    def __len__(self):
        return len(self.samples)


def _get_default_impute_fill_values(data_sources: Iterable[str]):
    fill_values = {}
    for source in data_sources:
        if source.startswith("omics_"):
            fill_values[source] = False
        else:
            fill_values[source] = -1

    return fill_values


def _get_default_impute_dtypes(data_sources: Iterable[str]):
    dtypes = {}
    for source in data_sources:
        if source.startswith("omics_"):
            dtypes[source] = torch.bool
        else:
            dtypes[source] = torch.float

    return dtypes


def prepare_one_hot_omics_data(
    genotype_array: torch.Tensor,
    na_augment_perc: float,
    na_augment_prob: float,
) -> torch.BoolTensor:
    array_bool = genotype_array.to(dtype=torch.bool)

    if na_augment_perc > 0 and na_augment_prob > 0:
        array_bool = make_random_omics_columns_missing(
            omics_array=array_bool,
            percentage=na_augment_perc,
            probability=na_augment_prob,
        )

    return array_bool


class DiskDataset(DatasetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.target_labels_dict:
            self.init_label_attributes()

        self.samples = self.set_up_samples()
        self.check_samples()

    # Note that dataloaders automatically convert arrays to tensors here, for labels
    def __getitem__(self, index: int) -> al_getitem_return:
        sample = self.samples[index]

        inputs_prepared = copy(sample.inputs)
        inputs_prepared = prepare_inputs_disk(
            inputs=inputs_prepared,
            na_augment_perc=self.na_augment[0],
            na_augment_prob=self.na_augment[1],
        )

        inputs_final = impute_missing_modalities_wrapper(
            inputs=inputs_prepared, data_dimensions=self.data_dimensions
        )

        target_labels = sample.target_labels
        sample_id = sample.sample_id

        return inputs_final, target_labels, sample_id

    def __len__(self):
        return len(self.samples)


def prepare_inputs_disk(
    inputs: Dict[str, Any], na_augment_perc: float, na_augment_prob: float
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}
    for name, data in inputs.items():

        if name.startswith("omics_"):
            array_raw = _load_one_hot_array(path=data)
            array_prepared = prepare_one_hot_omics_data(
                genotype_array=array_raw,
                na_augment_perc=na_augment_perc,
                na_augment_prob=na_augment_prob,
            )
            prepared_inputs[name] = array_prepared

        else:
            prepared_inputs[name] = inputs[name]

    return prepared_inputs


def prepare_inputs_memory(
    inputs: Dict[str, Any], na_augment_perc: float, na_augment_prob: float
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}
    for name, data in inputs.items():

        if name.startswith("omics_"):
            array_raw = data
            array_prepared = prepare_one_hot_omics_data(
                genotype_array=array_raw,
                na_augment_perc=na_augment_perc,
                na_augment_prob=na_augment_prob,
            )
            prepared_inputs[name] = array_prepared

        else:
            prepared_inputs[name] = inputs[name]

    return prepared_inputs


def impute_missing_modalities_wrapper(
    inputs: Dict[str, Any], data_dimensions: Dict[str, "DataDimensions"]
):
    impute_dtypes = _get_default_impute_dtypes(data_sources=data_dimensions.keys())
    impute_fill_values = _get_default_impute_fill_values(
        data_sources=data_dimensions.keys()
    )
    inputs_imputed = impute_missing_modalities(
        inputs=inputs,
        data_dimensions=data_dimensions,
        fill_values=impute_fill_values,
        dtypes=impute_dtypes,
    )

    return inputs_imputed


def impute_missing_modalities(
    inputs: Dict[str, Any],
    data_dimensions: Dict[str, "DataDimensions"],
    fill_values: Dict[str, Any],
    dtypes: Dict[str, Any],
) -> Dict[str, torch.Tensor]:

    for name, dimensions in data_dimensions.items():
        if name not in inputs:
            fill_value = fill_values[name]
            dtype = dtypes[name]
            shape = dimensions.channels, dimensions.height, dimensions.width

            imputed_tensor = impute_single_missing_modality(
                shape=shape, fill_value=fill_value, dtype=dtype
            )

            inputs[name] = imputed_tensor

    return inputs


def impute_single_missing_modality(shape: Tuple[int, ...], fill_value: Any, dtype: Any):
    imputed_tensor = torch.empty(shape, dtype=dtype).fill_(fill_value)
    return imputed_tensor


def _load_one_hot_array(path: Path) -> torch.Tensor:
    genotype_array_raw = np.load(str(path))
    genotype_array_raw = torch.from_numpy(genotype_array_raw).unsqueeze(0)

    return genotype_array_raw
